package Easysim::XML;

use strict;
use Easysim::XML::Document;
use Easysim::XML::Element;

################################################################################
# ($version,$encoding)=get_xml_info($message);
#
# Gets a version number an character set from not-parsed XML string.
# $message - input string containing XML data
# $version - XML version
# $encoding - character set
#
sub get_xml_info
{
	my ($msg)=@_;
	my ($str)=($$msg=~/^(<\?xml\s+.*?\?>)/);
	$$msg=~s/^<\?xml\s+.*?\?>\s*//s;
	my @a=($str=~/\sversion=(\"|')(\d+\.\d+)(\"|')(\s|\?>)/);
	my @b=($str=~/\sencoding=(\"|')([-\w]+)(\"|')(\s|\?>)/);
	die "Error: XML version not specified.\n" if(not @a or $a[0] ne $a[2]);
	die "Error: XML encoding not specified.\n" if(not @b or $b[0] ne $b[2]);
	return($a[1],$b[1]);
}

################################################################################
# ($final,$element)=get_xml_node($message,$dom);
#
# This function recursively parse the input XML string, each time bites a
# piece from the original message.
# $final - 1 if $element contains the resulted DOM element
# $element - the final DOM element
# $message - a XML string, probably not all of original already
# $dom - the DOM object used to create DOM elements
#
sub get_xml_node
{
	my ($msg,$dom)=@_;
	no warnings;
	$$msg=~s/^<([^>]*)>([^<]*)//s;
	my $text=$2;
	if($1){
		# some new node
		my ($name,$attrs)=($1=~/^(\/?\w+)\s*(.*)/s);
		if($name=~s/^\/(\w+)$/$1/s){
			$$msg="</$name>$$msg";
			return(0);
		}elsif(not $name=~/^\w+$/s){
			die "Error: bad tag: <$name>\n";
		}
		my ($final)=($attrs=~s/\/$//s);
		my $element=$dom->createElement($name);
		
		# get all attributes
		my @attr=split(/\s+/,$attrs);
		my %check;
		foreach (@attr){
			my @a=(/^(\w+)=(\"|')([^\"']*)(\"|')/);
			die "Bad attribute: $_\n" if(not @a or $a[1] ne $a[3]);
			die "Duplicit attribute: $_\n" if(defined $check{$a[0]});
			$element->setAttribute($a[0],$a[2]);
			$check{$a[0]}=1;
		}
		
		# text content
		$element->appendTextNode($text);

		return(1,$element) if($final);# the tag like <tag/>
		my ($ret,$child);
		do{
			# find all child nodes
			($ret,$child)=get_xml_node($msg,$dom);
			if($ret){
				$element->appendChild($child);
			}
		}while($ret);
		$$msg=~s/^[^<]*<([^>]*)>//s;
		my ($name2,$attrs2)=($1=~/^(\/?\w+)(.*)/s);
		if($name2=~s/^\/(\w+)$/$1/){
			die "Closing tag <$name2> doesn't match opening tag <$name>.\n" if($name ne $name2);
			return(1,$element);
		}
	}
}

################################################################################
# $parser=new();
#
# Creates new XML parser.
#
sub new
{
	my ($package)=@_;
	my $parser={};
	bless($parser);
	return($parser);
}

################################################################################
# $dom=$parser->parse_string($message);
#
# Creates DOM object from input message.
# $dom - the output DOM object
# $parser - the object of Easysim::XML used to parse strings
# $message - input XML message
#
sub parse_string
{
	my ($parser,$msg)=@_;
	my ($version,$encoding)=get_xml_info(\$msg);
	my $dom=Easysim::XML::Document->new($version,$encoding);

	$msg=~s/^([^<]*)//s;
	my $test=$1;
	if($test=~/\S/s){
		die "Error: parse_string: not a tag: '$test'\n";
	}
	if($msg=~/</){
		# the XML string isn't empty
		my ($ret,$node)=get_xml_node(\$msg,$dom);
		$dom->setDocumentElement($node);
	}
	return($dom);
}

1;

