#!/usr/bin/perl

use strict;
use Easysim::Communication;
use Easysim::Message;
use Easysim::Modules;
use Easysim::Utils;
use Easysim::XML;

# immediatelly print everything
STDOUT->autoflush(1);
STDERR->autoflush(1);

# come configuration...
my %config=(
	"port-gui"		=>2467,
	"port-compiler"		=>2468,
	"port-simulator"	=>2469,
	"module-compiler"	=>"Easysim::Compiler",
	"module-simulator"	=>"Easysim::Simulator",
	"module-checker"	=>"module_check",
	"module-entry"		=>"module_entry",
);
my $config=\%config;

# create end points for modules
my %end_point=(
	"gui"		=>new Easysim::Communication("server",$config->{'port-gui'}),
	"compiler"	=>new Easysim::Communication("server",$config->{'port-compiler'}),
	"simulator"	=>new Easysim::Communication("server",$config->{'port-simulator'}),
);

my @project;# the array containing all projects
my (@queue_compiler,@queue_simulator);# arrays containing new (unregistered) modules

# the kernel's main loop
while(1){
	# look for new connections
	check_ep_gui(\@project,$end_point{'gui'});
	check_ep_compiler(\@queue_compiler,$end_point{'compiler'});
	check_ep_simulator(\@queue_simulator,$end_point{'simulator'});

	# look for new messages
	check_msgs(\@project,\@queue_compiler,\@queue_simulator);

	Easysim::Utils::microsleep(0.25);
}

################################################################################
# check_ep_compiler(\@queue,$end_point);
#
# Looks for new connected compilers on $end_point. If found, inserts it into
# @queue.
#
sub check_ep_compiler
{
	my ($queue,$ep)=@_;
	my $channel=$ep->accept_connection();
	return unless(defined($channel));
	print "Kernel: New compiler: $channel\n";
	push @$queue,$channel;
}

################################################################################
# check_ep_gui(\@project,$end_point);
#
# Looks for new connected GUIs on $end_point. If found, creates new project
# in array @project.
#
sub check_ep_gui
{
	my ($project,$ep)=@_;
	my $channel=$ep->accept_connection();
	return unless(defined($channel));
	
	my $i=0;
	$i++ while(defined($project->[$i]));
	$project->[$i]={
		"id"		=>time."-$i",
		"channel-gui"	=>$channel,
		"stat"		=>"",
	};
	print "Kernel: New GUI: $channel\n";
}

################################################################################
# check_ep_simulato4(\@queue,$end_point);
#
# Looks for new connected simulators on $end_point. If found, inserts it into
# @queue.
#
sub check_ep_simulator
{
	my ($queue,$ep)=@_;
	my $channel=$ep->accept_connection();
	return unless(defined($channel));
	print "Kernel: New simulator: $channel\n";
	push @$queue,$channel;
}

################################################################################
# check_msg(\@project,\@queue_compiler,\@queue_simulator);
#
# This function looks for new messages from GUIs, compilers and simulators.
# If new message occured, calls appropriate function to proceed this message.
#
sub check_msgs
{
	my ($project,$queue_compiler,$queue_simulator)=@_;
	# proceed messages from GUIs
	for(my $i=0;$i<@$project;$i++){
		next unless(defined $project->[$i] and defined $project->[$i]{'channel-gui'});
		# messages <--- GUI
		my $channel_gui=$project->[$i]{'channel-gui'};
		my $dom=$channel_gui->receive_message();
		if(defined $dom and not $dom){
			# closed GUI's socket - shut down whole project
			print "Kernel: GUI $project->[$i]{'channel-gui'} disconnected\n";
			$project->[$i]=undef;
			next;
		}elsif(defined $dom){
			# some message <--- GUI
			proceed_msg_gui($project->[$i],$dom);
		}else{
			# messages ---> GUI
			proceed_msg_gui($project->[$i]);
		}

		if(defined($project->[$i]{'channel-compiler'})){
			# messages <--- compiler
			my $channel_compiler=$project->[$i]{'channel-compiler'};
			my $dom=$channel_compiler->receive_message();
			if(defined $dom and not $dom){
				# closed compiler's socket 
				print "Kernel: Compiler: Disconnected...\n";
				$project->[$i]{'channel-compiler'}=undef;
				next;
			}elsif(defined $dom){
				# some message <--- compiler
				proceed_msg_compiler($project->[$i],$dom);
			}else{
				# messages ---> compiler
				proceed_msg_compiler($project->[$i]);
			}
		}

		if(defined($project->[$i]{'channel-simulator'})){
			# messages <--- simulator
			my $channel_simulator=$project->[$i]{'channel-simulator'};
			my $dom=$channel_simulator->receive_message();
			if(defined $dom and not $dom){
				# closed simulator's socket 
				print "Kernel: Simulator: Disconnected...\n";
				$project->[$i]{'channel-simulator'}=undef;
				next;
			}elsif(defined $dom){
				# some message <--- simulator
				proceed_msg_simulator($project->[$i],$dom);
			}else{
				# messages ---> simulator
				proceed_msg_simulator($project->[$i]);
			}
		}
	}
	# proceed messages from new connected compilers
	my @queue_compiler_tmp;
	for(my $i=0;$i<scalar(@$queue_compiler);$i++){
		my $channel=$queue_compiler->[$i];
		my $doma=$channel->receive_message();
		if(defined $doma and not $doma){
			# closet compiler's socket - remove from queue
			print "Kernel: New compiler $channel lost...\n";
			next;
		}elsif(defined $doma){
			# some message!
			my $id=proceed_msg_compiler_new($channel,$doma);
			next unless($id);# not required message
			my ($j)=($id=~/\d+-(\d+)/);
			$project->[$i]{'channel-compiler'}=$channel;
			print "Kernel: Compiler: Registerd...\n";
			next;
		}
		push @queue_compiler_tmp,$channel;
	}
	@$queue_compiler=@queue_compiler_tmp;

	# proceed messages from new connected simulators
	my @queue_simulator_tmp;
	for(my $i=0;$i<scalar(@$queue_simulator);$i++){
		my $channel=$queue_simulator->[$i];
		my $doma=$channel->receive_message();
		if(defined $doma and not $doma){
			# closed simulator's socket - remove from queue
			print "Kernel: New simulator $channel lost...\n";
			next;
		}elsif(defined $doma){
			# some message!
			my $id=proceed_msg_simulator_new($channel,$doma);
			next unless($id);# not required message
			my ($j)=($id=~/\d+-(\d+)/);
			$project->[$i]{'channel-simulator'}=$channel;
			print "Kernel: Simulator: Registerd...\n";
			next;
		}
		push @queue_simulator_tmp,$channel;
	}
	@$queue_simulator=@queue_simulator_tmp;
}

################################################################################
# msg_gui_compile($project,$dom);
#
# This function creates new compiler or if simulator exists, sends all messages
# from GUI to it.
#
sub msg_gui_compile
{
	my ($proj,$dom)=@_;
	
	if($proj->{'channel-compiler'}){
		$proj->{'queue-compiler'}=Easysim::XML::Element->new('queue') if(not defined($proj->{'queue-compiler'}));
		my $queue_compiler=$proj->{'queue-compiler'};
		my $root=$dom->getDocumentElement;
		my @commands=$root->findnodes('/easysim/project/commands/command');
		for my $cmd (@commands){
			$queue_compiler->appendChild($cmd);
		}
	}else{
		Easysim::Modules::run_compiler($config,$proj,$dom);
	}
}

################################################################################
# msg_gui_newProject($project,$dom);
#
# This function retuns the project ID (if not sent yet) or creates new project
# for current GUI.
#
sub msg_gui_newProject
{
	my ($proj)=@_;

	my $msg=qq~<?xml version="1.0" encoding="iso8859-2"?>
		<easysim>
		    <project id="$proj->{'id'}">
		        <messages>
		            <kernel>
		                <msg type="status" status="success" command="newProject"/>
		            </kernel>
		        </messages>
		    </project>
		</easysim>
	~;
	my $a_parser=Easysim::XML->new();
	my $a_dom=$a_parser->parse_string("$msg");
	$proj->{'channel-gui'}->send_message($a_dom);
}

################################################################################
# msg_gui_simulate($project,$dom);
#
# This function creates new simulator or if simulator exists, sends all messages
# from GUI to it.
#
sub msg_gui_simulate
{
	my ($proj,$dom)=@_;
	
	if($proj->{'channel-simulator'}){
		$proj->{'queue-simulator'}=Easysim::XML::Element->new('queue') if(not defined($proj->{'queue-simulator'}));
		my $queue_simulator=$proj->{'queue-simulator'};
		my $root=$dom->getDocumentElement;
		my @commands=$root->findnodes('/easysim/project/commands/command');
		for my $cmd (@commands){
			$queue_simulator->appendChild($cmd);
		}
	}else{
		Easysim::Modules::run_simulator($config,$proj,$dom);
	}
}

################################################################################
# proceed_msg_compiler($project,$dom);
#
# This function proceeds all messages from and to compiler. If new message from
# compiler occured, calls appropriate functions.
# If in message queue for compiler contains a message, sends it to compiler.
#
sub proceed_msg_compiler
{
	my ($proj,$dom)=@_;

	if(defined $dom){
		# the queues for messages
		$proj->{'queue-gui'}=Easysim::XML::Element->new('queue') if(not defined($proj->{'queue-gui'}));
		$proj->{'queue-simulator'}=Easysim::XML::Element->new('queue') if(not defined($proj->{'queue-simulator'}));
		my ($queue_gui,$queue_simulator)=($proj->{'queue-gui'},$proj->{'queue-simulator'});
		if(not $queue_gui->findnodes('compiler')){
			my $tmp=Easysim::XML::Element->new('compiler');
			$queue_gui->appendChild($tmp);
			$queue_gui=$tmp;
		}
		if(not $queue_simulator->findnodes('compiler')){
			my $tmp=Easysim::XML::Element->new('compiler');
			$queue_simulator->appendChild($tmp);
			$queue_simulator=$tmp;
		}

		# proceed all messages
		my $root=$dom->getDocumentElement;
		my @msgs=$root->findnodes('/easysim/compiler/msg');
		for my $msg (@msgs){
			my $type=$msg->getAttribute("type");
			if($type eq "status"){
				$queue_gui->appendChild($msg);
			}elsif($type eq "signals"){
				$queue_gui->appendChild($msg);
			}elsif($type eq "vhdl"){
				$queue_simulator->appendChild($msg);
			}else{
				print "Kernel: Unknown message type: '$type'\n";
			}
		}
	}else{
		# messages ---> GUI
		return unless(defined $proj->{'queue-compiler'});
		my @msgs=$proj->{'queue-compiler'}->getChildNodes();
		if(@msgs){
			# create a message for compiler
			my $msg=qq~<?xml version="1.1" encoding="iso8859-2"?>
				<easysim></easysim>
			~;
			my $parser=Easysim::XML->new();
			my $answer_dom=$parser->parse_string("$msg");
			my $answer=$answer_dom->getDocumentElement;
			($answer)=$answer->findnodes('/easysim');
			for my $msg (@msgs){
				$answer->appendChild($msg);
			}
			$proj->{'channel-compiler'}->send_message($answer_dom);
			my $queue=Easysim::XML::Element->new('queue');
			$proj->{'queue-compiler'}=$queue;
		}
	}
}

################################################################################
# $id=proceed_msg_compiler_new($channel,$dom);
#
# This function is called, when a new created compiler wants to be registered.
# If the call was correct, the compiler is registered, a confirming message is 
# sent back to it and this function returns the project ID.
#
sub proceed_msg_compiler_new
{
	my ($channel,$dom)=@_;
	my $root=$dom->getDocumentElement;
	my @commands=$root->findnodes('/easysim/project/commands/command');
	return(undef) if(scalar(@commands) != 1);
	return(undef) unless($commands[0]->getAttribute("name") eq "registerCompiler");
	my ($project)=$root->findnodes('/easysim/project');
	my $id=$project->getAttribute("id");
	return(undef) unless($id);
	my $msg=qq~<?xml version="1.0" encoding="iso8859-2"?>
		<easysim><project id="$id"><messages><kernel>
		<msg type="status" status="success" command="registerCompiler"/>
		</kernel></messages></project></easysim>
	~;
	my $parser=Easysim::XML->new();
	my $a_dom=$parser->parse_string("$msg");
	$channel->send_message($a_dom);
	return($id);
}

################################################################################
# proceed_msg_simulator($project,$dom);
#
# This function proceeds all messages from and to simulator. If new message from
# simulator occured, calls appropriate functions.
# If in message queue for simulator contains a message, sends it to simulator.
#
sub proceed_msg_simulator
{
	my ($proj,$dom)=@_;
	if(defined $dom){
		# the queues for messages
		$proj->{'queue-gui'}=Easysim::XML::Element->new('queue') if(not defined($proj->{'queue-gui'}));
		my $queue_gui=$proj->{'queue-gui'};
		if(not $queue_gui->findnodes('simulator')){
			my $tmp=Easysim::XML::Element->new('simulator');
			$queue_gui->appendChild($tmp);
			$queue_gui=$tmp;
		}

		# proceed all messages
		my $root=$dom->getDocumentElement;
		my @msgs=$root->findnodes('/easysim/simulator/msg');
		for my $msg (@msgs){
			my $type=$msg->getAttribute("type");
			if($type eq "status"){
				$queue_gui->appendChild($msg);
			}elsif($type eq "waveform"){
				$queue_gui->appendChild($msg);
			}else{
				print "Kernel: Unknown message type: '$type'\n";
			}
		}
	}else{
		# messages ---> simulator
		return unless(defined $proj->{'queue-simulator'});
		my @msgs=$proj->{'queue-simulator'}->getChildNodes();
		if(@msgs){
			# create a message for simulator
			my $msg=qq~<?xml version="1.1" encoding="iso8859-2"?>
				<easysim></easysim>
			~;
			my $parser=Easysim::XML->new();
			my $answer_dom=$parser->parse_string("$msg");
			my $answer=$answer_dom->getDocumentElement;
			($answer)=$answer->findnodes('/easysim');
			for my $msg (@msgs){
				$answer->appendChild($msg);
			}
			$proj->{'channel-simulator'}->send_message($answer_dom);
			my $queue=Easysim::XML::Element->new('queue');
			$proj->{'queue-simulator'}=$queue;
		}
	}
}

################################################################################
# $id=proceed_msg_simulator_new($channel,$dom);
#
# This function is called, when a new created simulator wants to be registered.
# If the call was correct, the simulator is registered, a confirming message is 
# sent to it and this function returns the project ID.
#
sub proceed_msg_simulator_new
{
	my ($channel,$dom)=@_;
	my $root=$dom->getDocumentElement;
	my @commands=$root->findnodes('/easysim/project/commands/command');
	return(undef) if(scalar(@commands) != 1);
	return(undef) unless($commands[0]->getAttribute("name") eq "registerSimulator");
	my ($project)=$root->findnodes('/easysim/project');
	my $id=$project->getAttribute("id");
	return(undef) unless($id);
	my $msg=qq~<?xml version="1.0" encoding="iso8859-2"?>
		<easysim><project id="$id"><messages><kernel>
		<msg type="status" status="success" command="registerSimulator"/>
		</kernel></messages></project></easysim>
	~;
	my $parser=Easysim::XML->new();
	my $a_dom=$parser->parse_string("$msg");
	$channel->send_message($a_dom);
	return($id);
}

################################################################################
# proceed_msg_gui($project,$dom);
#
# This function proceeds all messages from and to GUI. If new message from GUI
# occured, calls appropriate functions.
# If in message queue for GUI contains a message, sends it to GUI.
#
sub proceed_msg_gui
{
	my ($proj,$dom)=@_;
	if(defined $dom){
		# messages <--- GUI
		my $root=$dom->getDocumentElement;
		my @commands=$root->findnodes('/easysim/project/commands/command');

		# first messages from GUI
		for my $cmd (@commands){
			my $name=$cmd->getAttribute("name");
			print "Kernel: GUI: Command '$name'\n";
			if($name eq "newProject"){
				msg_gui_newProject($proj);
			}elsif($name eq "compile"){
				msg_gui_compile($proj,$dom);
			}elsif($name eq "simulate"){
				msg_gui_simulate($proj,$dom);
			}
		}
	}else{
		# messages ---> GUI
		return unless(defined $proj->{'queue-gui'});
		my @msgs=$proj->{'queue-gui'}->getChildNodes();
		if(@msgs){
			# create a message for GUI
			my $msg=qq~<?xml version="1.1" encoding="iso8859-2"?>
				<easysim><project id="$proj->{'id'}"><messages/>
				</project></easysim>
			~;
			my $parser=Easysim::XML->new();
			my $answer_dom=$parser->parse_string("$msg");
			my $answer=$answer_dom->getDocumentElement;
			($answer)=$answer->findnodes('/easysim/project/messages');
			for my $msg (@msgs){
				$answer->appendChild($msg);
			}
			$proj->{'channel-gui'}->send_message($answer_dom);
			my $queue=Easysim::XML::Element->new('queue');
			$proj->{'queue-gui'}=$queue;
		}
	}
}

